/* milter-callback - a sendmail milter to perform variuos sender verifications
 Copyright (C) 2008  Eugene M. Zheganin, milter-callback@norma.perm.ru

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/   

#include <milter-callback.h>
#include <commonprocs.h>
#include <dealwithmxes.h>

extern char *logString;

int
processMXpart(char *formattedEmail, char *mailHost, struct mlfiPriv *privData) {
    int returnCode, returnSubCode;

    returnCode = dealWithMXes(formattedEmail, mailHost, privData);
    snprintf(logString, MAXLOGLINE, "dealWithMXes() returned %d.", returnCode);
    switch (returnCode) {
        case 0:
    	    /* 0 - all of the MXes accepted emailAddress. */
    	    snprintf(logString, MAXLOGLINE, "all of the MXes accepted mail to %s.", formattedEmail);
    	    logNow(logString, 2, privData -> threadCounter);
    	    return CB_PASS;
    	    break;
	case 1:
	    /* 1 - one of the MXes accepted emailAddress. */
	    snprintf(logString, MAXLOGLINE, "some of the MXes accepted mail to %s.", formattedEmail);
	    logNow(logString, 2, privData -> threadCounter);
	    return CB_PASS;
	    break;
	case 2:
	    /* 2 - no MX for emailAddress found. */
	    /* need to check emailAddress on host and/or on sending host */
	    snprintf(logString, MAXLOGLINE, "no MX records for the hostpart of %s found, need direct processing on the target host.", formattedEmail);
	    logNow(logString, 2, privData -> threadCounter);
	    returnSubCode = checkServerResponse(formattedEmail, mailHost, NULL, 0, privData);
	    snprintf(logString, MAXLOGLINE, "checkServerResponse() returned %d.", returnSubCode);
	    logNow(logString, 2, privData -> threadCounter);
	    switch (returnSubCode) {
	        case 0:
	    	    /* 0 - email accepped */
	    	    snprintf(logString, MAXLOGLINE, "Got relay accepting message to %s.", formattedEmail);
	    	    logNow(logString, 2, privData -> threadCounter);
	    	    return(CB_PASS);
		    break;
		case 1:
		    /* 1 - email not accepted */
		    snprintf(logString, MAXLOGLINE, "Got relay rejecting message to %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_REJECT);
	    	    break;
		case 2:
		    /* 2 - email tempfailed */
		    snprintf(logString, MAXLOGLINE, "Got relay tempfailing message to %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_TEMPFAIL);
		    break;
		case 3:
		    /* 3 - socket error */
		    snprintf(logString, MAXLOGLINE, "Got a socket error, returning relay tempfailing message to %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_TEMPFAIL);
		    break;
		case 4:
		    /* 4 - DNS error */
		    snprintf(logString, MAXLOGLINE, "Got a DNS error, returning relay tempfailing message from %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_TEMPFAIL);
		    break;
		case 5:
		    /* 5 - unknown answer code */
		    snprintf(logString, MAXLOGLINE, "Got an unknown SMTP answer code, returning relay tempfailing message from %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_TEMPFAIL);
		    break;
		case 6:
		default:
		    /* 6 - memory error */
		    snprintf(logString, MAXLOGLINE, "Got a memory or some unknown error, returning relay tempfailing message from %s.", formattedEmail);
		    logNow(logString, 2, privData -> threadCounter);
		    return(CB_TEMPFAIL);
		    break;
	    }
	    break;
	case 3:
	default:
	    /* 3 - some software/memory error happened. */
	    snprintf(logString, MAXLOGLINE, "Got a memory or connection error while parsing message from %s.", formattedEmail);
	    logNow(logString, 2, privData -> threadCounter);
	    return(CB_TEMPFAIL);
	    break;
	case 4:
	    /* 4 - none of the MXes accepted message. */
	    snprintf(logString, MAXLOGLINE, "All MXes rejected message from %s.", formattedEmail);
	    logNow(logString, 2, privData -> threadCounter);
	    return(CB_REJECT);
	    break;
	case 5:
	    /* 5 - MXes mostly tempfailed message. */
	    snprintf(logString, MAXLOGLINE, "None of the MXes accepted message, some tempfailed message from %s.", formattedEmail);
	    logNow(logString, 2, privData -> threadCounter);
	    return(CB_TEMPFAIL);
	    break;
    }
    /* some unknown code found */
    snprintf(logString, MAXLOGLINE, "Unknown answer code found, while parsing message from %s.", formattedEmail);
    logNow(logString, 0, privData -> threadCounter);
    return(CB_TEMPFAIL);
}
